module 'emop.parser' [ParsedFileCacheRef, ParseResultRef, OperatorsIdRef, LevelsIdRef, FreshVarCounterRef] {
   constructor: unescaped/0,atom/4,atom/0,variable/0,doc/0,xfy/0,left/0,doc/3,op/6,number/3,some/3,callable/5,empty/2,end_of_line/1,escaped/0,atom/3,file/2,error/2,toplevel/0,eos/0,eol/0,yes/0,binary/0,right/0,unespaced/0,yfx/0,doc/4,error/1,precede/0, none/0, colon/3,comment/4,functor/4,close/3,end_of_file/1,catch/0,for/0,findall/0,bagof/0,setof/0,throw/0,true/0,fail/0,! / 0,repeat/0,this/0,xfx/0,fy/0,yf/0,fx/0,xf/0,succede/0,string/3,operator/0, 'module'/7, notation/3, clause/6.

  fresh_var(V) :-
     (
        data.var:check(FreshVarCounterRef)
     ->
        FreshVarCounterRef = data.ref.non_backtrackable:new(0)
     ;
        true
     ),
     V = 'var'(
        'data.atom':format(
           '_~d', [data.number.(FreshVarCounterRef++)]
        )
     ).

   eliminate_parenthesis('()'(V0, _, _), V1) :-
      !,
      eliminate_parenthesis(V0, V1).
   eliminate_parenthesis(V, V).

   is_variable(Atom, Escaped, Args) :-
      Args = none,
      (
         Escaped = unescaped
      ->
         data.atom:sub(Atom, 0, 1, _, FirstLetter),
         (
            data.term:(
               FirstLetter >= 'A',
               FirstLetter =< 'Z'
            )
         ;
            FirstLetter = '_'
         ),
         !
      ;
         Escaped = variable
      ).


   fresh_operator_id(V) :-
     (
        data.var:check(OperatorsIdRef)
     ->
        OperatorsIdRef = data.ref.non_backtrackable:new(0)
     ;
        true
     ),
     V = data.number.(OperatorsIdRef++).

   fresh_level_id(V) :-
     (
        data.var:check(LevelsIdRef)
     ->
        LevelsIdRef = data.ref.non_backtrackable:new(0)
     ;
        true
     ),
     V = data.number.(atom(LevelsIdRef++)).

   parse_file(Filename, Stream, EmopSgn, Ast) :-
      ParsedFileCache = ParsedFileCacheRef:get,
      ChangedDate = sys.file:last_modification(Filename),
      (
         data.list:member('cache'(Filename, ParsedDate, Ast0), ParsedFileCache),
         data.term:(ChangedDate =< ParsedDate)
      ->
         Ast = Ast0
      ;
         (
            data.var:check(ParseResultRef)
         ->
            ParseResultRef = data.ref.non_backtrackable:new([])
         ;
            true
         ),
         (
            (
               ParseResultRef:set(parse_stream(Filename, 'file', Stream, EmopSgn))
            ->
               fail
            ;
               throw('error'('parsing_error'))
            )
         ;
            Ast0 = ParseResultRef:get,
            ParseResultRef:set([]),
            ParsedFileCacheRef:set(['cache'(Filename, ChangedDate, Ast0) | ParsedFileCache]),
            Ast = Ast0
         )
      ).

   parse_stream(Filename, Mode, Stream, EmopSgn, Ast) :-
      (
         Lexems = emop.lexer:utf8_stream(Stream, Filename)
      ->
         true
      ;
         throw(error('lexing_error'))
      ),
      parse(Lexems, Mode, EmopSgn, Ast0),
      (
         Ast0 = callable('.', 'operator', [Ast, empty(Pb, Pe)], _, _)
      ->
         (
            Mode = 'toplevel'
         ->
            'emop.message':warning(
               'Superfluous dot at the end of the goal',
               file(Pb, Pe)
            )
         ;
            'emop.message':warning(
               'Superfluous dot at the end of the file',
               file(Pb, Pe)
            )
         )
      ;
         Ast = Ast0
      ).

   parse(Lexems, Mode, EmopSgn, Ast) :-
      module Parser [Lexems, Mode, Parser, EmopSgn] {
         parse_left(Delim, Ops, Values, Ast) :-
            Lexems:read(Lexem),
            Parser:parse_left(Lexem, Delim, Ops, Values, Ast).

         parse_right(Delim, Ops, Values, Ast) :-
            Lexems:read(Lexem),
            Parser:parse_right(Lexem, Delim, Ops, Values, Ast).

         parse_left(
            '.('(Pb, Pe),
            _Delim,
            _Ops,
            _Values,
            _Ast
         ) :-
            throw(error('Scope delimiter should be preceded by a module', file(Pb, Pe))).

         parse_left(
            comment(Kind, Comment, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            (
               Kind = doc
            ->
               Ops0 = [op(1250, xfy, left, doc(Comment, PB, PE), PB, PE) | Ops]
            ;
               Ops0 = Ops
            ),
            Parser:parse_left(Delim, Ops0, Values, Ast).

         parse_left(
            number(Number, PB, Pos_end),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            Values0 = [number(Number, PB, Pos_end) | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).

         parse_left('('(PB, PE), Delim, Ops, Values, Ast) :-
            Parser:parse_left(some(')', PB, PE), [], [], V),
            Values0 = ['()'(V, PB, PE) | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).

         parse_left('['(PB, PE), Delim, Ops, Values, Ast) :-
            Parser:parse_left(some(']', PB, PE), [], [], Contents),
            (
               Contents = callable('|', 'operator', [Seq, Tail], _, _)
            ->
               true
            ;
               Seq = Contents,
               Tail = callable('[]', unescaped, none, PB, PE)
            ),
            Parser:sequence_list(Seq, List),
            Parser:append(List, Tail, PB, PE, V),
            Values0 = [V | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).
         
         parse_left('{'(PB, PE), Delim, Ops, Values, Ast) :-
            Parser:parse_left(some('}', PB, PE), [], [], V),
            (
               V = empty(_, _)
            ->
               Args = []
            ;
               Args = [V]
            ),
            Values0 = [callable('{}', unescaped, Args, PB, PE) | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).

         parse_left(
            atom(Atom, Escaped, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            (
               Escaped = unescaped,
               Parser:(
                  op(Atom, TrueAtom, Priority, Assoc),
                  op_left(Assoc)
               )
            ->
               Lexems:read(Lexem),
               (
                  Atom = '-',
                  Lexem = number(Number, _, PE0)
               ->
                  Values0 = [number('data.number':(-Number), PB, PE0) | Values],
                  Parser:parse_right(Delim, Ops, Values0, Ast)
               ;  
                  Op = op(Priority, Assoc, left, TrueAtom, PB, PE),
                  Ops0 = [Op | Ops],
                  Parser:parse_left(Lexem, Delim, Ops0, Values, Ast)
               )
            ;
               Escaped = unescaped,
               Atom = 'module'
            ->
               parse_module(PB, PE, Delim, Ops, Values, Ast)
            ;
               Escaped = unescaped
            ->
               Value = callable(Atom, unescaped, none, PB, PE),
               Parser:parse_right(Delim, Ops, [Value | Values], Ast)
            ;
               Value = callable(Atom, Escaped, none, PB, PE),
               Parser:parse_right(Delim, Ops, [Value | Values], Ast)
            ).

         parse_left(string(String, PB, PE), Delim, Ops, Values, Ast) :-
            Parser:parse_string(String, PB, PE, Delim, Ops, Values, Ast).

         parse_string(String, PB, PE, Delim, Ops, Values, Ast) :-
            Lexems:read(Lexem),
            (
               Lexem = end_of_line(_)
            ->
               Parser:parse_string(String, PB, PE, Delim, Ops, Values, Ast)
            ;
               Lexem = string(String0, _, PE0)
            ->
               Parser:parse_string(
                  'data.list':append(String, String0),
                  PB,
                  PE0,
                  Delim,
                  Ops,
                  Values,
                  Ast
               )
            ;
               read_string(String, PB, PE, Value),
               Parser:parse_right(Lexem, Delim, Ops, [Value | Values], Ast)
            ).

         parse_module(PB, PE, Delim, Ops, Values, Ast) :-
            Lexems:read(Lexem),
            (
               Lexem = atom(_Atom, _Escaped, _AtomPB, _AtomPE)
            ->
               Name = Lexem,
               Lexems:read(Lexem0)
            ;
               Name = atom('_', unescaped, PB, PE),
               Lexem0 = Lexem
            ),
            (
               parse_module_env(Lexem0, Curly_begin, Curly_end, Env)
            ->
               true
            ;
               throw(error('Module body is missing', file(PB, PE)))
            ),
            parse_module_body(Curly_begin, Curly_end, Env, Name, PB, PE, V),
            Values0 = [V | Values],
            parse_right(Delim, Ops, Values0, Ast).

         parse_module_env(Lexem, Curly_begin, Curly_end, Env) :-
            (
               Lexem = '{'(Curly_begin, Curly_end)
            ->
               Env = 'none'
            ;
               Lexem = '['(Env_begin, Env_end)
            ->
               Parser:parse_left(some(']', Env_begin, Env_end), [], [], Seq),
               Parser:sequence_list(Seq, Env),
               Lexems:read(Lexem0),
               (
                  Lexem0 = '{'(Curly_begin, Curly_end)
               ->
                  true
               ;
                  fail
               )
            ;
               fail
            ).

         parse_module_body(Curly_begin, Curly_end, Env, Name, PB, PE, V) :-
            Parser:parse_left(some('}', Curly_begin, Curly_end), [], [], Body),
            (
               data.var:check(EmopSgn)
            ->
               throw(error('Unallowed module', file(PB, PE)))
            ;
               Clauses = EmopSgn:enum_clauses(Body),
               Sgn = EmopSgn:extract_sgn(Clauses)
            ->
               true
            ;
               throw('sgn error')
            ),
            V = module(Name, Env, Body, Sgn, Clauses, PB, Curly_end).

         parse_left(
            functor(Atom, Escaped, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            Parser:parse_left(some(')', PB, PE), [], [], Seq),
            Parser:sequence_list(Seq, List),
            Value = callable(Atom, Escaped, List, PB, PE),
            Values0 = [Value | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).

         parse_left(
            close(Symbol, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            Values0 = [empty(PB, PE) | Values],
            Parser:check_delim(Delim, Symbol, PB, PE),
            Parser:flush(Ops, Values0, Ast).

         parse_left(end_of_file(P), Delim, Ops, Values, Ast) :-
            parse_end_of_file(Delim, Ops, [empty(P, P) | Values], Ast).

         parse_left(end_of_line(_), Delim, Ops, Values, Ast) :-
            Parser:parse_left(Delim, Ops, Values, Ast).

         parse_right(
            comment(Kind, _, Pb, Pe),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            !,
            (
               Kind = 'doc'
            ->
               throw(error('Documentation not allowed here', file(Pb, Pe)))
            ;
               true
            ),
            Parser:parse_right(Delim, Ops, Values, Ast).

         parse_right(
            close(Symbol, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            !,
            Parser:check_delim(Delim, Symbol, PB, PE),
            Parser:flush(Ops, Values, Ast).

         parse_right(end_of_file(_), Delim, Ops, Values, Ast) :-
            !,
            parse_end_of_file(Delim, Ops, Values, Ast).

         parse_end_of_file(Delim, Ops, Values, Ast) :-
            catch(
               (
                  Parser:check_no_delim(Delim),
                  Parser:flush(Ops, Values, Ast)
               ),
               E,
               (
                  (
                     Mode = 'toplevel'
                  ->
                     fail
                  ;
                     throw(E)
                  )
               )
            ).

         parse_right(end_of_line(_), Delim, Ops, Values, Ast) :-
            !,
            Parser:parse_right(Delim, Ops, Values, Ast).

         parse_right(
            atom('|', unescaped, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            Delim = some(']', _, _),
            !,
            Parser:flush_upto(
               Ops,
               Values,
               PE,
               1050,
               'xfx',
               Ops0,
               Values0
            ),
            Op = op(1050, 'xfx', binary, '|', PB, PE),
            Ops1 = [Op | Ops0],
            Parser:parse_left(Delim, Ops1, Values0, Ast).

         parse_right(
            atom(Atom, Escaped, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            (
               Escaped = unescaped
            ->
               true
            ),
            (
               Parser:op(Atom, TrueAtom, Priority, Assoc),
               (
                  Parser:op_binary(Assoc)
               ->
                  Kind = binary
               ;
                  Parser:op_right(Assoc)
               ->
                  Kind = right
               )
            ->
               true
            ),
            !,
            Parser:flush_upto(
               Ops,
               Values,
               PE,
               Priority,
               Assoc,
               Ops0,
               Values0
            ),
            (
               Kind = binary
            ->
               Op = op(Priority, Assoc, binary, TrueAtom, PB, PE),
               Ops1 = [Op | Ops0],
               Parser:parse_left(Delim, Ops1, Values0, Ast)
            ;
               Values0 = [Arg | Tail],
               Values1 = [callable(TrueAtom, unespaced, [Arg], PB, PE) | Tail],
               Parser:parse_right(Delim, Ops0, Values1, Ast)
            ).

         parse_right(Lexem, Delim, Ops, Values, Ast) :-
            (
               Lexem = '['(_PB, _PE)
            ;
               Lexem = '{'(_PB, _PE)
            ),
            Values = [Value | Values0],
            eliminate_parenthesis(Value, Value0),
            Value0 = callable(F, Escaped, Args, PB, PE),
            Args \= none,
            Name = atom('_', unescaped, PB, PE),
            parse_module_env(Lexem, Curly_begin, Curly_end, Env),
            !,
            parse_module_body(Curly_begin, Curly_end, Env, Name, PB, PE, V),
            'data.list':append(Args, [V], Args0),
            Values1 = [callable(F, Escaped, Args0, PB, PE) | Values0],
            Parser:parse_right(Delim, Ops, Values1, Ast).
            
         parse_right(
            '.('(Pb, Pe),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            !,
            Parser:flush_upto(
               Ops,
               Values,
               Pe,
               80,
               xfy,
               Ops0,
               [PrefixArg | Values0]
            ),
            Parser:parse_left(some(')', Pb, Pe), [], [], V),
            Values1 = ['.()'(PrefixArg, V, Pb, Pe) | Values0],
            Parser:parse_right(Delim, Ops0, Values1, Ast).

         parse_right(Term, Delim, Ops, Values, Ast) :-
            Values = [First | _],
            emop.parse:position(First, file(_, Pe)),
            Parser:flush_upto(
               Ops,
               Values,
               Pe,
               100,
               xfy,
               Ops0,
               Values0
            ),
            Ops1 = [op(100, xfy, binary, '$~$', Pe, Pe) | Ops0],
            Parser:parse_left(Term, Delim, Ops1, Values0, Ast).

         sequence_list(empty(_, _), []) :-
            !.
         sequence_list(callable(',', 'operator', [Left, Right], _, _), [Left | Tail]) :-
            !,
            Parser:sequence_list(Right, Tail).
         sequence_list(Term, [Term]).

         append([], Tail, _, _, Tail).
         append([H | T], Tail, PB, PE, L) :-
            L = callable('.', escaped, [H, L0], PB, PE),
            Parser:append(T, Tail, PB, PE, L0).

         read_string([], PB, PE, callable('[]', escaped, [], PB, PE)).
         read_string([H | T], PB, PE, Term) :-
            Term = callable(
               '.',
               escaped,
               [callable(H, escaped, [], PB, PE), Tail],
               PB,
               PE
            ),
            Parser:read_string(T, PB, PE, Tail).

         op('.', '.', 1300, xfy).
         op(':-', ':-', 1200, xfx).
         op(',', ',', 1000, xfy).
         op('=', '=', 700, xfy).
         op(':', ':', 50, yfx).

         op_left(fx).
         op_left(fy).

         op_binary(xfx).
         op_binary(xfy).
         op_binary(yfx).

         op_right(xf).
         op_right(yf).

         relative(xfy, xfy, succede).
         relative(yfx, yfx, precede).

         check_no_delim(none).
         check_no_delim(some(_, PB, PE)) :-
            throw(error('Unmatched delimiter', file(PB, PE))).

         check_delim(none, _, PB, PE) :-
            throw(error('Unmatched delimiter', file(PB, PE))).
         check_delim(some(Delim, _, _), Delim, _, _) :-
            !.
         check_delim(some(_, PB, _), _, _, PE) :-
            throw(
               error('Mismatched delimiters', file(PB, PE))
            ).

         apply_op(left, Atom, PB, PE, Values0, [V | Values1]) :-
            (
               Values0 = [Arg0 | Values1]
            ->
               (
                  Atom = doc(Comment, PB0, PE0)
               ->
                  V = doc(Comment, PB0, PE0, Arg0)
               ;
                  V = callable(Atom, 'operator', [Arg0], PB, PE)
               )
            ;
               throw(
                  error('Missing operand', file(PB, PE))
               )
            ).
         apply_op(binary, Atom, PB, PE, Values0, [V | Values1]) :-
            (
               Values0 = [Arg1, Arg0 | Values1]
            ->
               V = callable(Atom, 'operator', [Arg0, Arg1], PB, PE)
            ;
               throw(
                  error('Missing right operand', file(PB, PE))
               )
            ).

         flush([], Values, Ast) :-
            (
               [Ast] = Values
            ->
               true
            ;
               throw(error('flush'))
            ).
         flush(
            [doc(Comment, PB, PE) | Tail],
            Values0,
            Ast
         ) :-
            Values0 = [Value | Values1],
            Parser:flush(Tail, [doc(Comment, PB, PE, Value) | Values1], Ast).
         flush(
            [op(_, _, Kind, Op, PB, PE) | Tail],
            Values0,
            Ast
         ) :-
            Parser:(
               apply_op(Kind, Op, PB, PE, Values0, Values1),
               flush(Tail, Values1, Ast)
            ).

         flush_upto([], Values, _, _, _, [], Values).
         flush_upto(
            [Op | Tail],
            Values0,
            PE,
            Priority1,
            Assoc1,
            Ops1,
            Values1
         ) :-
            (
               Op = op(Priority0, Assoc0, Kind, Atom, PB, _),
               (
                  'data.number':(Priority0 < Priority1)
               ;
                  Priority0 = Priority1,
                  (
                     Parser:relative(Assoc0, Assoc1, Rel)
                  ->
                     (
                        Rel = precede
                     ->
                        true
                     ;
                        fail
                     )
                  ;
                     (
                        Atom = ':-'
                     ->
                        throw(error('Operator :- is not associative.  A \'.\' is probably missing.', file(PB, PE)))
                     ;
                        throw(error(data.atom:format('Operator conflict with ~a', [Atom]), file(PB, PE)))
                     )
                  )
               )
            ->
               Parser:(
                  apply_op(Kind, Atom, PB, PE, Values0, Values),
                  flush_upto(
                     Tail,
                     Values,
                     PE,
                     Priority1,
                     Assoc1,
                     Ops1,
                     Values1
                  )
               )
            ;
               Ops1 = [Op | Tail],
               Values1 = Values0
            ).
      },
      Parser:parse_left(none, [], [], Ast).

   internal_reparse(Ast, Grammars) =
      reparse_flat(emop.parse:reflat(Ast), Grammars).

   reparse_flat(Flat, Grammars) =
      remove_concat(Flat, Grammars).

   remove_concat(Ast0, Grammars, Ast1) :-
      module [Grammars] {
         do(callable('$~$', operator, [A, B], _, _), Ast) :-
            !,
            List = [A | emop.parse:enum_concat_list(B)],
            do(reparse_list(List, Grammars), Ast).
         do(callable(F, Escaped, List, Pb, Pe), Ast) :-
            !,
            Ast = callable(F, Escaped, do_list(List), Pb, Pe).
         do('()'(SubTerm, Pb, Pe), Ast) :-
            !,
            Ast = '()'(do(SubTerm), Pb, Pe).
         do('.()'(A, B, Pb, Pe), Ast) :-
            !,
            Ast = '.()'(do(A), B, Pb, Pe).
         do(module(Name, Env, Body, Sgn, Clauses, Pb, Pe), Ast) :-
            !,
            Ast = module(Name, do_list(Env), Body, Sgn, Clauses, Pb, Pe).
         do(Ast, Ast).

         do_list(none) = ! >> none.
         do_list(List) = data.list:iter(List, this).
      }:do(Ast0, Ast1).

   constructor: grammar/5.

   grammar_empty = grammar([], [], data.record:{}, [], data.record:{}).

   reparse_list(List, Grammars, Ast) :-
      module [Grammars] {
         prefix_op('fx').
         prefix_op('fy').

         postfix_op('xf').
         postfix_op('yf').

         binary_op('xfx').
         binary_op('xfy').
         binary_op('yfx').

         constructor: op/5, priority/2, operator/5.

         reassoc_left([H | T], OpStack, ValueStack, Ast) :-
            T \= [],
            H = callable(Op, 'unescaped', none, Pb, Pe),
            data.list:member(Grammar, Grammars),
            Grammar match grammar(
               data.list:member(operator(Op, OpId, Alias, 'prefix', Assoc)),
               _, _, _, _
            ),
            !,
            reassoc_left(T, ['prefix'(Op, OpId, Alias, Assoc, Pb, Pe) | OpStack], ValueStack, Ast).
         reassoc_left([H | T], OpStack, ValueStack) = (
            reassoc_right(T, OpStack, [H | ValueStack])
         ).

         apply_op('binary'(_Op, _OpId, Alias, _Assoc, Pb, Pe), [Right, Left | ValueStack]) =
            [callable(Alias, 'operator', [Left, Right], Pb, Pe) | ValueStack].
         apply_op('prefix'(_Op, _OpId, Alias, _Assoc, Pb, Pe), [Left | ValueStack]) =
            [callable(Alias, 'operator', [Left], Pb, Pe) | ValueStack].

         flush([], [Result]) = Result.
         flush([OpRepr | Ops], ValueStack) = flush(Ops, apply_op(OpRepr, ValueStack)).
            !.

         less_priority(Op, _, _, Op).
         less_priority(Op0, Cache, Prefix, Op1) :-
            data.list:member(Grammar, Prefix),
            Grammar match grammar(
               _,
               data.list:member(priority(Op0, Op2)),
               _, _, _
            ),
            \+ data.list:member(Op2, Cache),
            less_priority(Op2, [Op2 | Cache], Prefix, Op1).

         compare_priorities(Op0, Op1, Cmp) :-
            data.list:append(Prefix, _, Grammars),
            (
               less_priority(Op0, [Op0], Prefix, Op1)
            ->
               (
                  less_priority(Op1, [Op1], Prefix, Op0)
               ->
                  Cmp = '='
               ;
                  Cmp = '<'
               )
            ;
               less_priority(Op1, [], Prefix, Op0)
            ->
               Cmp = '>'
            ;
               fail
            ),
            !.

         flush_upto([], ValueStack, _OpInfo, [], ValueStack).
         flush_upto([OpRepr | Ops0], ValueStack0, OpInfo, Ops1, ValueStack1) :-
            (
               OpRepr = 'binary'(Op, OpId, Alias, Assoc, Pb, _Pe)
            ;
               OpRepr = 'prefix'(Op, OpId, Alias, Assoc, Pb, _Pe)
            ),
            !,
            OpInfo = op(Op0, OpId0, Assoc0, _Pb0, Pe0),
            (
               OpId = OpId0
            ->
               (
                  (Assoc = 'yfx'; Assoc = 'yf')
               ->
                  Action = 'flush'
               ;
                  Assoc = 'xfy'
               ->
                  Action = 'stop'
               ;
                  throw(error(data.atom:format('Operator ~a is not associative.', [Op0, Op]), file(Pb, Pe0)))
               )
            ;
               compare_priorities(OpId, OpId0) match (
                  '='
               ->
                  (
                     Assoc = 'yfx',
                     Assoc0 = ('yfx'; 'yf')
                  ->
                     Action = 'flush'
                  ;
                     Assoc = ('xfy'; 'fy'),
                     Assoc0 = 'xfy'
                  ->
                     Action = 'stop'
                  ;
                     Assoc = 'xfx',
                     Assoc0 = 'xfx'
                  ->
                     throw(error(data.atom:format('Operators ~a and ~a are not associative', [Op0, Op]), file(Pb, Pe0)))
                  ;
                     Assoc = 'xfx'
                  ->
                     throw(error(data.atom:format('Operator ~a is not associative', [Op]), file(Pb, Pe0)))
                  ;
                     Assoc0 = 'xfx'
                  ->
                     throw(error(data.atom:format('Operator ~a is not associative', [Op0]), file(Pb, Pe0)))
                  ;
                     (
                        Assoc = 'xfy', RightOp = Op,
                        Assoc0 = 'yfx', LeftOp = Op0
                     ;
                        Assoc = 'yfx', LeftOp = Op,
                        Assoc0 = 'xfy', RightOp = Op0
                     )
                  ->
                     throw(error(data.atom:format('Operator ~a is left associative whereas operator ~a is right associative', [LeftOp, RightOp]), file(Pb, Pe0)))
                  )
               ;
                  '<'
               ->
                  Action = 'stop'
               ;
                  '>'
               ->
                  Action = 'flush'
               )
            ->
               true
            ;
               throw(error(data.atom:format('Operators ~a and ~a are not compatible', [Op, Op0]), file(Pb, Pe0)))
            ),
            (
               Action = 'flush'
            ->
               flush_upto(Ops0, apply_op(OpRepr, ValueStack0), OpInfo, Ops1, ValueStack1)
            ;
               Ops1 = [OpRepr | Ops0],
               ValueStack1 = ValueStack0
            ).

         check_postfix(Op, Grammars, OpId, Alias, Assoc) :-
            data.list:member(Grammar, Grammars),
            Grammar match grammar(
               data.list:member(operator(Op, OpId, Alias, 'postfix', Assoc)),
               _, _, _, _
            ),
            !.

         check_infix(Op, Grammars, OpId, Alias, Assoc) :-
            data.list:member(Grammar, Grammars),
            Grammar match grammar(
               data.list:member(operator(Op, OpId, Alias, 'infix', Assoc)),
               _, _, _, _
            ),
            !.

         check_postfix_or_infix(Op, Grammars) :-
            data.list:member(Grammar, Grammars),
            Grammar match grammar(
               data.list:member(operator(Op, _, _, Kind, _)),
               _, _, _, _
            ),
            Kind = ('postfix'; 'infix'),
            !.

         reassoc_right([], OpStack, ValueStack) = (
            flush(OpStack, ValueStack)
         ).
         reassoc_right([H | T], OpStack, ValueStack, Ast) :-
            H = '.()'(Sub, Pb, Pe),
            check_postfix('.()', Grammars, OpId, _Alias, Assoc),
            !,
            flush_upto(OpStack, ValueStack, op('.()', OpId, Assoc, Pb, Pe), OpStack0, ValueStack0),
            ValueStack0 = [Arg | ValueStack1],
            Value = '.()'(Arg, Sub, Pb, Pe),
            reassoc_right(T, OpStack0, [Value | ValueStack1], Ast).
         reassoc_right([H | T], OpStack, ValueStack, Ast) :-
            H = callable(Op, 'unescaped', none, Pb, Pe),
            check_postfix(Op, Grammars, OpId, Alias, Assoc),
            (
               check_infix(Op, Grammars, _, _, _)
            ->
               (
                  T = []
               ->
                  true
               ;
                  T = [TH | _],
                  TH = callable(TOp, 'unescaped', none, _, _)
               ->
                  check_postfix_or_infix(TOp, Grammars)
               )
            ;
               true
            ),
            !,
            flush_upto(OpStack, ValueStack, op(Op, OpId, Assoc, Pb, Pe), OpStack0, ValueStack0),
            ValueStack0 = [Arg | ValueStack1],
            Value = callable(Alias, 'operator', [Arg], Pb, Pe),
            reassoc_right(T, OpStack0, [Value | ValueStack1], Ast).
         reassoc_right([H | T], OpStack, ValueStack, Ast) :-
            (
               H = callable(Op, 'unescaped', none, Pb, Pe),
               T \= [],
               Tail = T
            ;
               Op = '',
               Tail = [H | T],
               ValueStack = [First | _],
               file(_, Pe) = emop.parse:position(First),
               Pb = Pe
            ),
            check_infix(Op, Grammars, OpId, Alias, Assoc),
            !,
            flush_upto(OpStack, ValueStack, op(Op, OpId, Assoc, Pb, Pe), OpStack0, ValueStack0),
            reassoc_left(Tail, ['binary'(Op, OpId, Alias, Assoc, Pb, Pe) | OpStack0], ValueStack0, Ast).
         reassoc_right([H | T], _ValueStack, _OpStack, _Ast) :-
            (
               T = []
            ->
               throw(error('Postfix operator expected.', emop.parse:position(H)))
            ;
               throw(error('Infix or postfix operator expected.', emop.parse:position(H)))
            ).
      }:reassoc_left(List, [], [], Ast).


         symbol_list(List, SymbolList) :-
            module [] {
               do([], []).
               do(
                  [
                     callable(Symbol, _, none, _, _),
                     callable('as', 'unescaped', none, _, _),
                     callable(Alias, _, none, _, _)
                  |
                     Tail
                  ],
                  List
               ) :-
                  !,
                  List = [(Symbol, Alias) | do(Tail)].
               do([callable(Symbol, _, none, _, _) | Tail], List) :-
                  List = [(Symbol, Symbol) | do(Tail)].
            }:do(SymbolList, List).

         find_operator('operator'(F, Symbol, Pb, Pe), GrammarRef, Grammars, OperatorId) :-
            (
               (
                  Grammar = GrammarRef:get
               ;
                  data.list:member(Grammar, Grammars)
               ),
               Grammar match grammar(
                  data.list:member('operator'(Symbol, OperatorId, _, F, _)),
                  _, _, _, _
               )
            ->
               true
            ;
               data.var:check(F)
            ->
               throw(
                  error(
                     data.atom:format('Unknown operator ~a', [Symbol]),
                     file(Pb, Pe)
                  )
               )
            ;
               F = 'infix'
            ->
               throw(
                  error(
                     data.atom:format('Unknown infix operator ~a', [Symbol]),
                     file(Pb, Pe)
                  )
               )
            ;
               F = 'prefix'
            ->
               throw(
                  error(
                     data.atom:format('Unknown prefix operator ~a', [Symbol]),
                     file(Pb, Pe)
                  )
               )
            ;
               F = 'postfix'
            ->
               throw(
                  error(
                     data.atom:format('Unknown postfix operator ~a', [Symbol]),
                     file(Pb, Pe)
                  )
               )
            ).

         constructor: priority/2, level/4, syntax/3, level/3.

         expand_priorities(Id, F, List, Pb, Pe, GrammarRef, Grammars) :-
            (
               read_symbol_definition(List, Other, Tail)
            ->
               true
            ;
               throw(
                  error(
                     'Operator name expected',
                     file(Pb, Pe)
                  )
               )
            ),
            Id0 = find_operator(Other, GrammarRef, Grammars),
            GrammarRef:set(
               GrammarRef:get match grammar(
                  _,
                  (Priorities -> [priority(Id, Id0) | Priorities]),
                  _, _, _
               )
            ),
            (
               F = '='
            ->
               GrammarRef:set(
                  GrammarRef:get match grammar(
                     _,
                     (Priorities' -> [priority(Id0, Id) | Priorities']),
                     _, _, _
                  )
               )
            ;
               true
            ),
            (
               Tail = []
            ->
               true
            ;
               Tail = [H | T]
            ->
               (
                  H = callable(F0, 'unescaped', none, _, _),
                  (F0 = '=' ; F0 = '<')
               ->
                  expand_priorities(Id0, F0, T, Pb, Pe, GrammarRef, Grammars)
               ;
                  throw(
                     error(
                        'Priority comparison < or = expected',
                        emop.parse:position(H)
                     )
                  )
               )
            ).

         expand_syntax(GrammarRef, Grammars, List, Pb, Pe) :-
            read_symbol_definition(List, First, [callable(F, 'unescaped', none, _, _) | Tail]),
            (F = '=' ; F = '<'),
            !,
            expand_priorities(find_operator(First, GrammarRef, Grammars), F, Tail, Pb, Pe, GrammarRef, Grammars).
         expand_syntax(GrammarRef, _Grammars, List, Pb, Pe) :-
            Assoc = (List match (
               [
                  callable('infix', 'unescaped', none, _, _),
                  callable('left', 'unescaped', none, _, _),
                  callable('associative', 'unescaped', none, _, _)
               |
                  symbol_list(Symbols)
               ]
            ->
               'yfx'
            ;
               [
                  callable('infix', 'unescaped', none, _, _),
                  callable('right', 'unescaped', none, _, _),
                  callable('associative', 'unescaped', none, _, _)
               |
                  symbol_list(Symbols)
               ]
            ->
               'xfy'
            ;
               [
                  callable('infix', 'unescaped', none, _, _)
               |
                  symbol_list(Symbols)
               ]
            ->
               'xfx'
            ;
               [
                  callable('prefix', 'unescaped', none, _, _)
               |
                  symbol_list(Symbols)
               ]
            ->
               'fy'
            ;
               [
                  callable('postfix', 'unescaped', none, _, _)
               |
                  symbol_list(Symbols)
               ]
            ->
               'yf'
            )),
            !,
            List = [callable(Kind, _, _, _, _) | _],
            data.list:iter(Symbols) [Pb, Pe, Kind, Assoc, GrammarRef] {
               do((Symbol, Alias)) :-
                  (
                     Symbol = '',
                     Kind \= 'infix'
                  ->
                     throw(error('The empty operator can only be infix.', file(Pb, Pe)))
                  ;
                     true
                  ),
                  GrammarRef:set(
                     GrammarRef:get match grammar(
                        (Operators ->
                           [
                              'operator'(
                                 Symbol, fresh_operator_id, Alias, Kind, Assoc
                              ) 
                           | 
                              Operators
                           ]),
                        _, _, _, _
                     )
                  )
            }.
         expand_syntax(_GrammarRef, _Grammars, _List, Pb, Pe) :-
            throw(
               error(
                  'Invalid syntax definition',
                  file(Pb, Pe)
               )
            ).

         read_symbol_definition(
            [
               callable(F, 'unescaped', none, _, _),
               callable(A, _, none, Pb, Pe)
            |
               Tail
            ],
            'operator'(F, A, Pb, Pe),
            Tail
         ) :-
            (F = 'prefix'; F = 'postfix'; F = 'infix'),
            !.
         read_symbol_definition(
            [
               callable(A, _, none, Pb, Pe)
            |
               Tail
            ],
            'operator'(_F, A, Pb, Pe),
            Tail
         ).

         extract_level_scope(Desc, Pb, Pe, LevelList, GrammarRef, Grammars, Desc0) :-
             (
                Desc = ['()'(Spec, _, _) | Desc0]
             ->
                SpecList = emop.parse:enum_concat_list(Spec),
                LevelList = module [GrammarRef, Grammars] {
                   do([]) = [].
                   do([callable(Level, _, none, _, _) | T0 ], [Level0 | T1]) :-
                      Level0 = find_level(GrammarRef, Grammars, Level),
                      T0 match (
                         [
                            callable(',', unescaped, none, _, _)
                            | T0s
                         ]
                      ->
                         do(T0s, T1)
                      ;
                         _
                      ->
                         do(T0, T1)
                      ).
                }:do(SpecList)
             ;
                throw(error('Missing level scope', file(Pb, Pe)))
             ).

         find_level(GrammarRef, Grammars, Level, Id) :-
            GrammarRef:get =
               grammar(_Operators, _Priorities, LevelIds, _Levels, _Notations),
            (
               Id = LevelIds:{Level},
               \+ data.var:check(Id)
            ->
               true
            ;
               data.list:member(Grammar, Grammars),
               Grammar = grammar(_, _, LevelIds', _, _),
               Id = LevelIds':{Level},
               \+ data.var:check(Id)
            ->
               true
            ;
               Id = fresh_level_id,
               LevelIds:{Level} = Id
            ).

         extract_level('()'(Desc, _, _), LevelList, GrammarRef, Grammars) = extract_level(Desc, LevelList, GrammarRef, Grammars).

         extract_level(callable(F, Escaped, Args, _, _), LevelList, GrammarRef, Grammars) =
            level(
               LevelList,
               F,
               Escaped,
               data.list:iter(Args) [GrammarRef, Grammars] {
                  do(callable(Level, _, none, _, _)) = ! >> find_level(GrammarRef, Grammars, Level).
                  do(callable('seq', unescaped, [callable(Level, _, none, _, _)], _, _)) = 'seq'(find_level(GrammarRef, Grammars, Level)).
                  do(callable('*', operator, [callable(Level, _, none, _, _)], _, _)) = 'seq'(find_level(GrammarRef, Grammars, Level)).
               }
            ).

         grammar_star(Grammars) =
            [grammar([operator('*', fresh_operator_id, '*', 'postfix', 'xf')], [], data.record:{}, [], data.record:{}) | Grammars].

         expand_level(GrammarRef, Grammars, Desc, Pb, Pe) :-
            extract_level_scope(Desc, Pb, Pe, LevelList, GrammarRef, Grammars, Desc0),
            Grammars0 = grammar_star([GrammarRef:get | Grammars]),
            Desc1 = reparse_flat(emop.parse:reconcat_list(Desc0, file(Pb, Pe)), Grammars0),
            (
               Level = extract_level(Desc1, LevelList, GrammarRef, Grammars)
            ->
               GrammarRef:set(
                  GrammarRef:get match grammar(
                     _, _, _,
                     (Levels -> [Level | Levels]),
                     _
                  )
               )
            ;
               throw(error('Invalid level specification', file(Pb, Pe)))
            ).

         constructor: operator/5.

         add_notation(GrammarRef, Grammars, Desc, Pb, Pe) :-
            extract_level_scope(Desc, Pb, Pe, LevelList, GrammarRef, Grammars, Desc0),
            (
               module [Tail, Pb, Pe] {
                  cut_guard([callable('->', unescaped, none, _, _) | Tail], []) :-
                     !.
                  cut_guard([A | B], [A | B0]) :-
                     cut_guard(B, B0).

                  cut([callable('->', unescaped, none, _, _) | Tail], [], [callable('true', unescaped, none, Pb, Pe)]) :-
                     !.
                  cut([callable('|', unescaped, none, _, _) | B], [], Guard) :-
                     !,
                     cut_guard(B, Guard).
                  cut([A | B], [A | B0], Guard) :-
                     cut(B, B0, Guard).
               }:cut(Desc0, Head, Guard)
            ->
               true
            ;
               throw(error('Notation without ->.', file(Pb, Pe)))
            ),
            Grammars1 = grammar_star([GrammarRef:get | Grammars]),
            Head0 = reparse_flat(emop.parse:reconcat_list(Head, file(Pb, Pe)), Grammars1),
            Guard0 = emop.parse:reparse_flat(
               emop.parse:reconcat_list(Guard, file(Pb, Pe))
            ) {
               syntax: infix left associative ';' ',' '+' '-' '*' '/' ':'.
               syntax: prefix '#'.
               syntax: postfix '*'.
               syntax: ';' < ',' < ':' < '+' = '-' < infix '*' = '/' < '#' < postfix '*'.
            },
            Tail0 = reparse_flat(emop.parse:reconcat_list(Tail, file(Pb, Pe)), Grammars1),
            Env = data.record:{},
            InferLengthRef = data.ref.backtrackable:new(),
            module [Env, InferLengthRef, Grammars1] {
               rename(Mode, Ast) = module [Mode, Env, InferLengthRef, Grammars1] {
                  get_fresh(F, Pb, Pe, X) :-
                     X = Env:{F},
                     (
                        data.var:check(X)
                     ->
                        true
                     ;
                        throw(error(data.atom:format('The variable ~a is already bound.', [F]), file(Pb, Pe)))
                     ).

                  do_term_variable(F, Pb, Pe, Kind, Result) :-
                     (
                        F = '_'
                     ->
                        Result = callable('_', variable, none, Pb, Pe)
                     ;
                     Mode match (
                        'head'
                     ->
                        X = get_fresh(F, Pb, Pe),
                        X = 'term'(V),
                        data.term:make(Kind, [V], Result)
                     ;
                        'tail'
                     ->
                        U = Env:{F},
                        U match (
                           data.var:check
                        ->
                           Result = 'fresh'(_),
                           U = 'term'(Result)
                        ;
                           ('term'(V); 'sequence'(V); 'sequence'(V, _); 'module env var'(V))
                        ->
                           Result = V
                        ;
                           'predicate'(_, _, _)
                        ->
                           throw(error(data.atom:format('The variable ~a should only appear as predicate name.', [F]), file(Pb, Pe)))
                        ;
                           'module body'(_)
                        ->
                           throw(error(data.atom:format('The variable ~a should only appear as module body.', [F]), file(Pb, Pe)))
                        ;
                           'module env seq'(_)
                        ->
                           throw(error(data.atom:format('The variable ~a should only appear as module environment.', [F]), file(Pb, Pe)))
                        )
                     ;
                        'arg'
                     ->
                        U = Env:{F},
                        U match (
                           data.var:check
                        ->
                           Result = 'fresh'(_),
                           U = 'term'(Result)
                        ;
                           ('term'(V); 'module env var'(V))
                        ->
                           Result = V
                        ;
                           _
                        ->
                           throw(error(data.atom:format('The variable ~a is not bound to a simple term and cannot appear as argument of a predicate.', [F]), file(Pb, Pe)))
                        )
                     ;
                        'seq'(AccuLists, AccuVars, LenRef)
                     ->
                        U = Env:{F},
                        U match (
                           data.var:check
                        ->
                           U = 'term'('fresh'(_))
                        ;
                           ('term'(V); 'module env var'(V))
                        ->
                           Result = V
                        ;
                           _
                        ->
                           do_len_seq(U, F, Pb, Pe, LenRef, AccuVars, AccuLists, Result)
                        )
                     )).

                  do_len_seq(U, F, Pb, Pe, LenRef, AccuVars, AccuLists, Result) :-
                     U match (
                        'sequence'(V)
                     ->
                        (
                           LenRef = F
                        ->
                           Result = X,
                           AccuVars:set([X | AccuVars:get]),
                           AccuLists:set([V | AccuLists:get])
                        ;
                           throw(error(data.atom:format('Sequences ~a and ~a do not necessarily have the same length.', [LenRef, F]), file(Pb, Pe)))   
                        )
                     ;
                        'sequence'(V, W)
                     ->
                        do_len_seq_ref(V, W, Pb, Pe, AccuLists, AccuVars, LenRef, Result)
                     ).

                  do_len_seq_ref(V, W, Pb, Pe, AccuLists, AccuVars, LenRef, Result) :-
                     (
                        LenRef = W
                     ->
                        Result = X,
                        AccuVars:set([X | AccuVars:get]),
                        AccuLists:set([V | AccuLists:get])
                     ;
                        throw(error(data.atom:format('Sequences ~a and ~a do not necessarily have the same length.', [LenRef, W]), file(Pb, Pe)))   
                     ).

                   do_module_body(Body0, Sgn0, Clauses0, Pb, Pe, Body1) :-
                      (
                         Body0 = callable(BodyF, BodyEscaped, BodyArgs, _, _),
                         is_variable(BodyF, BodyEscaped, BodyArgs)
                      ->
                         (
                            Mode = 'head'
                         ->
                            (
                               Env:{BodyF} = 'module body'(V)
                            ->
                               Body1 = V
                            ;
                               throw(error(data.atom:format('The variable ~a should only bind a module body.', [BodyF]), file(Pb, Pe)))
                            )
                         ;
                            U = Env:{BodyF},
                            (
                               data.var:check(U)
                            ->
                               throw(error(data.atom:format('Fresh variables cannot appear as module body.', [BodyF]), file(Pb, Pe)))
                            ;
                               true
                            ),
                            (
                               U = 'module body'(V)
                            ->
                               Body1 = V
                            ;
                               throw(error(data.atom:format('The variable ~a should be a module body.', [BodyF]), file(Pb, Pe)))
                            )
                         )
                      ;
                         Mode = 'tail'
                      ->
                         Rewrite = this,
                         Clauses1 = data.list:iter(Clauses0) [Rewrite, Grammars1] {
                            do(clause(F, Args, B, Doc, Pb, Pe), Clause1) :-
                               (
                                  Args = none
                               ->
                                  Args1 = none
                               ;
                                  Args1 = data.list:iter(Args) {
                                     do(Arg) = rename('arg', internal_reparse(Arg, Grammars1)).
                                  }
                               ),
                               B1 = Rewrite:do(internal_reparse(B, Grammars1)),
                               Clause1 = clause(F, Args1, B1, Doc, Pb, Pe).
                         },
                         Body1 = 'custom'(Sgn0, Clauses1)
                      ;
                         throw(error('Only a variable is allowed as module body in head of notations.', file(Pb, Pe)))
                      ).

                   head_seq(Args, Pb, Pe, Result) :-
                      (
                         Args =
                         [callable(V, VEscaped, VArgs, _, _)],
                         is_variable(V, VEscaped, VArgs)
                      ->
                         true
                      ;
                         throw(error('Sequence constructor should have a variable as unique argument', file(Pb, Pe)))
                      ),
                      (
                         Env:{V} = 'sequence'(X)
                      ->
                         Result = 'seq'(X)
                      ;
                         throw(error(data.atom:format('The variable ~a should only bind a sequence.', [V]), file(Pb, Pe)))
                      ).

                   tail_seq(Args, Pb, Pe, Result) :-
                      (
                         Args =
                         [
                            callable(V, VEscaped, VArgs, _, _),
                            callable(W, VEscaped, VArgs, _, _)
                         ],
                         is_variable(V, VEscaped, VArgs),
                         is_variable(W, VEscaped, VArgs)
                      ->
                         V0 = get_fresh(V, Pb, Pe),
                         Env:{V} = V0,
                         Env:{W} = W0,
                         (
                            data.var:check(W0)
                         ->
                            throw(error(data.atom:format('The variable ~a should be bound to a sequence.', [W]), file(Pb, Pe)))
                         ;
                            (
                               W0 = 'sequence'(X),
                               W = W1
                            ;
                               W0 = 'sequence'(_, W1)
                            )
                         ->
                            true
                         ;
                            throw(error(data.atom:format('The variable ~a is not bound to a sequence.', [W]), file(Pb, Pe)))
                         ),
                         V0 = 'sequence'(Y, W1),
                         Result = 'seq'(Y, X)
                      ;
                         Args = [Arg]
                      ->
                         (
                            Arg = callable(V, VEscaped, VArgs, _, _),
                            is_variable(V, VEscaped, VArgs)
                         ->
                            V0 = Env:{V},
                            (
                               data.var:check(V0)
                            ->
                               InferLengthRef:set(['infer'(V, W, X, Pb, Pe) | InferLengthRef:get]),
                               (
                                  Mode = 'seq'(AccuLists, AccuVars, LenRef)
                               ->
                                  V0 = 'sequence'(V, W),
                                  do_len_seq_ref(V, W, Pb, Pe, AccuLists, AccuVars, LenRef, Result)
                               ;
                                  V0 = 'sequence'(Y, W),
                                  Result = 'seq'(Y, X)
                               )
                            ;
                               (
                                  Mode = 'seq'(AccuLists, AccuVars, LenRef)
                               ->
                                  (
                                     do_len_seq(V0, V, Pb, Pe, LenRef, AccuVars, AccuLists, Result)
                                  ->
                                     true
                                  ;
                                     throw(error(data.atom:format('The variable ~a should be bound to a sequence', [V]), file(Pb, Pe)))
                                  )
                               ;
                                  (
                                     (
                                        V0 = 'sequence'(Se)
                                     ;
                                        V0 = 'sequence'(Se, _)
                                     )
                                  ->
                                     Result = Se
                                  ;
                                     throw(error(data.atom:format('The variable ~a should be bound to a sequence', [V]), file(Pb, Pe)))
                                  )
                               )
                            )
                         ;
                            AccuLists = data.ref.backtrackable:new([]),
                            AccuVars = data.ref.backtrackable:new([]),
                            Seq = rename('seq'(AccuLists, AccuVars, LenRef), Arg),
                            (
                               data.var:check(LenRef)
                            ->
                               throw(error('The sequence expression should be built from at least one sequence', file(Pb, Pe)))
                            ;
                               true
                            ),
                            Result = 'seq'(AccuLists:get, AccuVars:get, Seq)
                         )
                      ;
                         throw(error(data.atom:format('Sequences should either be bound to a variable according to the length of another sequence or built by a sequence expression.', [V]), file(Pb, Pe)))
                      ).

                   do(module(NameAtom, Env0, Body0, Sgn0, Clauses0, Pb, Pe), Result) :-
                      !,
                      NameAtom = atom(Name0, Escaped, PbName, PeName),
                      (
                         is_variable(Name0, Escaped, none)
                      ->
                         (
                            Name0 = '_'
                         ->
                            Name1 = atom(Name0, Escaped, PbName, PeName)
                         ;
                            do_term_variable(Name0, PbName, PeName, 'variable', Name1)
                         )
                      ;
                         throw(error('Only variables are supported for module name in notation.', file(PbName, PeName)))
                      ),
                      (
                         Env0 = none
                      ->
                         Env1 = none
                      ;
                         Env1 = data.list:iter(Env0) [Env, Mode] {
                            do(callable(F, Escaped, Args, Pb, Pe), R) :-
                               (
                                  is_variable(F, Escaped, Args)
                               ->
                                  (
                                     Mode = 'head'
                                  ->
                                     (
                                        Env:{F} = 'module env var'(V)
                                     ->
                                        R = 'variable'(V)
                                     ;
                                        throw(error(data.atom:format('The variable ~a should only bind a module environment variable.', [F]), file(Pb, Pe)))
                                     )
                                  ;
                                     U = Env:{F},
                                     (
                                        data.var:check(U)
                                     ->
                                        U = 'module env var'('fresh'(_))
                                     ;
                                        true
                                     ),
                                     (
                                        U = ('module env var'(V); 'module env seq'(V); 'term'(V))
                                     ->
                                        R = V
                                     ;
                                        throw(error(data.atom:format('The variable ~a should be a module environment variable.', [F]), file(Pb, Pe)))
                                     )
                                  )
                               ;
                                  F = '*',
                                  Escaped = operator,
                                  Args = [_]
                               ->
                                  Mode match (
                                     'head'
                                  ->
                                     (
                                        Args =
                                        [callable(V, VEscaped, VArgs, _, _)],
                                        is_variable(V, VEscaped, VArgs)
                                     ->
                                        VE = get_fresh(V, Pb, Pe),
                                        VE = 'module env seq'(X),
                                        R = 'seq'(X)
                                     ;
                                        throw(error('Sequence constructor should have a variable as unique argument', file(Pb, Pe)))
                                     )
                                  ;
                                     (
                                        'tail'
                                     ;
                                        'seq'(_, _, _)
                                     )
                                  ->
                                     tail_seq(Args, Pb, Pe, R)
                                  ;
                                     'arg'
                                  ->
                                        throw(error('Sequences are not allowed as arguments of a predicate', file(Pb, Pe)))
                                  )
                               ).
                         }
                      ),
                      do_module_body(Body0, Sgn0, Clauses0, Pb, Pe, Body1),
                      Result = 'module'(Name1, Env1, Body1, Pb, Pe).
                   do(callable(F, Escaped, Args, Pb, Pe), Result) :-
                      !,
                      (
                         is_variable(F, Escaped, Args)
                      ->
                         do_term_variable(F, Pb, Pe, 'variable', Result)
                      ;
                         Escaped = unescaped,
                         Mode = 'head',
                         (
                            (
                               F = 'var'
                            ;
                               F = 'number'
                            )
                         ->
                            Args =
                            [callable(V, VEscaped, VArgs, _, _)],
                            is_variable(V, VEscaped, VArgs)
                         )
                      ->
                         do_term_variable(V, Pb, Pe, F, Result)
                      ;
                         Escaped = operator,
                         Mode = 'head',
                         F = '*'
                      ->
                         Result = head_seq(Args, Pb, Pe)
                      ;
                         Escaped = operator,
                         F = '*',
                         (
                            Mode = 'tail'
                         ;
                            Mode = 'seq'(_, _, _)
                         ;
                            Mode = 'arg'
                         ->
                            throw(error('Sequences are not allowed as arguments of a predicate', file(Pb, Pe)))
                         )
                      ->
                         Result = tail_seq(Args, Pb, Pe)
                      ;
                         (
                            Escaped = unescaped
                         ->
                            (
                               (
                                  Mode = 'head'
                               ->
                                  get_fresh(F, Pb, Pe, Fe),
                                  Fe = 'predicate'(X, FEscaped, FArity),
                                  F0 = 'variable'(X, FEscaped, FArity),
                                  Escaped0 = unescaped
                               ;
                                  (
                                     Mode = 'tail'
                                  ;
                                     Mode = 'arg'
                                  )
                               ->
                                  Env:{F} = U,
                                  (
                                     data.var:check(U)
                                  ->
                                     F = F0,
                                     Escaped = Escaped0
                                  ;
                                     U = 'predicate'(X, FEscaped, _),
                                     F0  = X,
                                     Escaped0 = FEscaped
                                  )
                               )
                            ->
                               true
                            ;
                               throw(error(data.atom:format('The variable ~a should only bind a predicate.', [F]), file(Pb, Pe)))
                            )
                         ;
                            Escaped = Escaped0,
                            F = F0
                         ),
                         (
                            Args = none
                         ->
                            Args0 = none
                         ;
                            Args0 = data.list:iter(Args, this)
                         ),
                         Result = callable(F0, Escaped0, Args0, Pb, Pe)
                      ).
                   do('()'(Sub, Pb, Pe)) = ! >> '()'(do(Sub), Pb, Pe).
                   do(A) = A
                }:do(Ast).
          }:(
             rename('head', Head0, Head1),
             rename('tail', Tail0, Tail1)
          ),
          module [Env] {
             make(callable('true', unescaped, none, _, _), 'true') :-
                !.

             make(callable(':', operator, [X, Kind], _, _), Guard) :-
                !,
                (
                   Kind = callable(Kind0, unescaped, none, _, _)
                ->
                   true
                ;
                   throw(error('Invalid kind', emop.parse:position(Kind)))
                ),
                Kind0 match (
                   ('var'; 'number')
                ->
                   Guard = 'type_term'(Kind0, get_variable_arg(X))
                ;
                   ('constructor'; 'predicate'; 'symbol')
                ->
                   Guard = 'sgn_check'(Kind0, get_functor(X))
                ;
                   _
                ->
                   throw(error('Unknown kind', emop.parse:position(Kind)))
                ).

             make('()'(Sub, _, _), Guard) :-
                !,
                make(Sub, Guard).

             make(callable(',', operator, [A, B], _, _), Guard) :-
                !,
                make(A, Guard0),
                make(B, Guard1),
                Guard = ','(Guard0, Guard1).

             make(callable(';', operator, [A, B], _, _), Guard) :-
                !,
                make(A, Guard0),
                make(B, Guard1),
                Guard = ';'(Guard0, Guard1).

             make(X, _) :-
                throw(error('Invalid guard', emop.parse:position(X))).

             get_number(X, Y) :-
                X match (
                   number(Value, _, _)
                ->
                   Y = 'number'(Value)
                ;
                   callable('#', operator, [Arg], _, _)
                ->
                   module [Y, Env] {
                      do(Arg) :-
                        ( 
                           Arg = callable('*', operator, [Sub], _, _)
                        ->
                           (
                              Sub = callable(V, VEscaped, VArgs, _, _),
                              is_variable(V, VEscaped, VArgs),
                              V0 = Env:{V},
                              \+ data.var:check(V0),
                              V0 = ('sequence'(S); 'sequence'(_, S))
                           ->
                              Y = 'length'(S)
                           ;
                              throw(error('Should be a variable bound to a sequence', emop.parse:position(Sub)))
                           )
                        ;
                           Arg = callable(V, VEscaped, VArgs, _, _),
                           V0 = Env:{V},
                           \+ data.var:check(V0)
                        ->
                           (
                              V0 = ('sequence'(S), 'sequence'(_, S))
                           ->
                              Y = 'length'(S)
                           ;
                              V0 = ('predicate'(_, _, Arity))
                           ->
                              Y = 'number'(Arity)
                           )
                        ;
                           Arg = '()'(Sub, _, _)
                        ->
                           do(Sub)
                        ;
                           throw(error('Should be a bound variable', emop.parse:position(Arg)))
                        ).
                   }:do(Arg)
                ;
                   callable('+', operator, [Left, Right], _, _)
                ->
                   Y = '+'(get_number(Left), get_number(Right))
                ;
                   callable('-', operator, [Left, Right], _, _)
                ->
                   Y = '-'(get_number(Left), get_number(Right))
                ;
                   callable('*', operator, [Left, Right], _, _)
                ->
                   Y = '*'(get_number(Left), get_number(Right))
                ;
                   callable('/', operator, [Left, Right], _, _)
                ->
                   Y = '/'(get_number(Left), get_number(Right))
                ;
                   '()'(Sub, _, _)
                ->
                   Y = get_number(Sub)
                ;
                   _
                ->
                   throw(error('Should be a number', emop.parse:position(X)))
                ).

             get_functor(X0, Y) :-
                X = eliminate_parenthesis(X0),
                (
                   X = callable('/', operator, [Functor0, Arity0], _, _)
                ->
                   Functor = eliminate_parenthesis(Functor0),
                   Arity = get_number(Arity0)
                ;
                   Functor = X
                ),
                (
                   Functor = callable(':', operator, [Mod0, Symbol0], _, _)
                ->
                   get_variable_arg(Mod0, Mod),
                   Symbol = eliminate_parenthesis(Symbol0)
                ;
                   Mod = 'scope',
                   Symbol = Functor
                ),
                (
                   Symbol = callable(F, unescaped, none, _, _)
                ->
                   (
                      Fe = Env:{F},
                      \+ data.var:check(Fe),
                      Fe = 'predicate'(Predicate, Escaped, FArity)
                   ->
                      (
                         data.var:check(Arity)
                      ->
                         Arity = 'number'(FArity)
                      ;
                         true
                      )
                   ;
                      throw(error(data.atom:format('~a should be bound to a predicate', [F]), emop.parse:position(Symbol)))
                   )
                ;
                   Symbol = callable(Predicate, Escaped, none, _, _)
                ->
                   (
                      data.var:check(Arity)
                   ->
                      throw(error('Missing arity', emop.parse:position(X)))
                   ;
                      true
                   )
                ;
                   throw(error('Should be a functor', emop.parse:position(Symbol)))
                ),
                Y = functor(Mod, Predicate, Escaped, Arity).

             get_variable_arg(X, Y) :-
                (
                   X = callable(V, VEscaped, VArgs, _, _),
                   is_variable(V, VEscaped, VArgs),
                   V0 = Env:{V},
                   \+ data.var:check(V0),
                   V0 = 'term'(Y)
                ->
                   true
                ;
                   throw(error('Should be a variable', emop.parse:position(X)))
                ).
                
          }:make(Guard0, Guard1),
          data.list:iter(InferLengthRef:get) [Env] {
             do('infer'(V, W, X, Pb, Pe)) :-
                (
                   data.var:check(W)
                ->
                   throw(error(data.atom:format('Cannot infer the length of ~a', [V]), file(Pb, Pe)))
                ;
                   Env:{W} = 'sequence'(X)
                ).
          },
          GrammarRef:get(grammar(_, _, _, _, Notations)),
          data.list:iter(LevelList) {
             do(Level) :-
                LevelNotations = Notations:{Level},
                (
                   data.var:check(LevelNotations)
                ->
                   LevelNotations = (data.record:{}, data.ref.backtrackable:new([]))
                ;
                   true
                ),
                LevelNotations = (Indexed, Unindexed),
                (
                   Unindexed:get([]),
                   Head1 = callable(F, _, Args, _, _),
                   data.atom:check(F),
                   \+ data.list:member('seq'(_), Args)
                ->
                   List = Indexed:{data.atom:format('~a/~n', [F, (Args = none -> 0; data.list:length(Args))])},
                   (
                      data.var:check(List)
                   ->
                      List = data.ref.backtrackable:new([])
                   ;
                      true
                   )
                ;
                   List = Unindexed
                ),
                List:set([notation(Head1, Guard1, Tail1) | List:get]).
          }.

      add_to_grammar(syntax(Desc, Pb, Pe), GrammarRef, SupGrammars) :-
         expand_syntax(GrammarRef, SupGrammars, Desc, Pb, Pe).
      add_to_grammar(level(Desc, Pb, Pe), GrammarRef, SupGrammars) :-
         expand_level(GrammarRef, SupGrammars, Desc, Pb, Pe).
      add_to_grammar(notation(Desc, Pb, Pe), GrammarRef, SupGrammars) :-
         add_notation(GrammarRef, SupGrammars, Desc, Pb, Pe).
      add_to_grammar('symbol'(_, _, _, _, _), _GrammarRef, _SupGrammars).

      make_grammar(Sgn, SupGrammars, Grammar) :-
         GrammarRef = data.ref.backtrackable:new(grammar_empty),
         data.list:iter(Sgn) [GrammarRef, SupGrammars] {
            do(Item) :-
               (
                  add_to_grammar(Item, GrammarRef, SupGrammars)
               ->
                  true
               ;
                  true
               ).
         },
         Grammar = GrammarRef:get.
},
ParsedFileCacheRef = data.ref.non_backtrackable:new([])