module emop.refl {
   name(M) = ModName :-
      data.term:functor(M, ModName, _).

   env(M) = Env :-
      data.term:decomposition(M, [ModName | ModArgs]),
      (
         ModArgs = [Env]
      ->
         true
      ;
         ModArgs = []
      ->
         top:g_read(ModName, Env)
      ).


   has_predicate(ModName, Env, Head, HasHead) :-
      data.atom:concat(ModName, ':$has_predicate', Has),
      HasHead = data.term:make(Has, [Head, Env]).

   has_dynamic_predicate(ModName, Env, Head, HasHead) :-
      data.atom:concat(ModName, ':$has_dynamic_predicate', Has),
      HasHead = data.term:make(Has, [Head, Env]).

   has(M, P) :-
      top:call(:has_predicate(:name(M), :env(M), P)).

   ensure_predicate(M, Head, ModF) :-
      data.term:functor(M, ModName, _),
      data.term:functor(Head, F, N),
      data.atom:concat(data.atom:concat(ModName, ':'), F, ModF),
      (
         has(M, Head)
      ->
         true
      ;
         data.term:functor(Proto, F, N),
         top:asserta(:has_dynamic_predicate(ModName, _, Proto) :- !),
         data.term:decomposition(Proto, [_ | Args]),
         top:asserta(data.term:make(ModName, [Proto, Pos, Env]) :- !, data.term:make(ModF, data.list:append(Args, [Pos, Env])))
      ).

   asserta(M, Env, Pos, Clause) :-
      decompose_clause(Clause, Head, Body),
      ensure_predicate(M, Head, ModF),
      top:asserta(:compose_module_clause(M, Env, Head, ModF, Pos, Body)).

   asserta(M, Env, Clause) :-
      asserta(M, Env, _Pos, Clause).

   assertz(M, Env, Pos, Clause) :-
      decompose_clause(Clause, Head, Body),
      ensure_predicate(M, Head, ModF),
      top:assertz(:compose_module_clause(M, Env, Head, ModF, Pos, Body)).

   assertz(M, Env, Clause) :-
      assertz(M, Env, _Pos, Clause).

   existence_error(ModName) =
      (data.term:make(ModName, [T, _, Pos]) :- functor(T, F, N), throw(error(existence_error(proceduce(F / N)), Pos))).

   fresh_module(Env, M) :-
      ModName = data.atom:new,
      M = data.term:make(ModName, [data.term:make(env, Env)]),
      ProtoM = data.term:make(ModName, [EnvVar]),
      HasPredicate = :has_predicate(ModName, Head0, Env0),
      HasDynamicPredicate = :has_dynamic_predicate(ModName, Head0, Env0),
      top:asserta(HasPredicate :- HasDynamicPredicate),
      top:(
         asserta(HasDynamicPredicate),
         retract(HasDynamicPredicate)
      ),
      top:asserta(dynamic_emop_call(ProtoM, Head, Pos) :- !, data.term:make(ModName, [Head, Pos, EnvVar])),
      top:asserta(:existence_error(ModName)).

   unpickle(M, ModVar) :-
      ModVar = :fresh_module([M]),
      data.term:functor(ModVar, ModName, _),
      (
         has(M, has(_))
      ->
         top:asserta(:has_dynamic_predicate(ModName, env(M0), Head) :- emop_call(M0, has(Head), unpickle))
      ;
         top:asserta(:has_dynamic_predicate(ModName, env(M0), _))
      ),
      top:asserta(data.term:make(ModName, [X, Pos, env(M0)]) :- !, emop_call(M0, do(X), Pos)).

   pickle(M, ModVar) :-
      ModVar = :fresh_module([M]),
      asserta(ModVar, [M0], Pos, (do(X) :- emop_call(M0, X, Pos))).

   compose_module_clause(M, Env, Head, ModF, Pos, Body, Clause) :-
      data.term:functor(:env(M), env, EnvSize),
      (
         data.list:length(Env) = EnvSize
      ->
         true
      ;
         throw(error(invalid_env_size, assertz))
      ),
      data.term:decomposition(Head, [_ | HeadArgs]),
      ModArg = data.term:make(ModF, data.list:append(HeadArgs, [Pos, data.term:make(env, Env)])),
      Clause = (ModArg :- Body).

   decompose_clause((Head :- Body), Head, Body) :- !.
   decompose_clause(Head, Head, true).
}
