module 'emop.writer' {
   constructor: callable/2, var/1, args/1, term/1, number/1.

   callable(Term, P) :-
      data.term:decomposition(Term, [F | Args]),
      P = callable(F, Args).

   condition(Cond, True, False) =
      callable(';', [callable('->', [Cond, True]), False]).

   sequence([]) = 'true'.
   sequence([H | T]) =
      'data.list':fold_left(T, H) [] {
         do(I, S) = callable(',', [S, I]).
      }.

   write(S, P) :-
      module Writer [S, Output_char, Output_arg] {
         write(var(Var)) :-
            !,
            S:print('_'),
            (
               Var \= '_'
            ->
               (
                  (
                     data.atom:sub(Var, _, 1, _, C),
                     (
                        data.term:(
                           (C >= 'a', C =< 'z');
                           (C >= 'A', C =< 'Z');
                           (C >= '0', C =< '9'))
                     ->
                        S:print(C)
                     ;
                        C = '_'
                     ->
                        S:print('__')
                     ;
                        S:print('_'),
                        S:print(data.char:code(C))
                     ),
                     fail
                  ;
                     true
                  )
               ->
                  true
               )
            ;
               true
            ).
         write(callable(F, Args)) :-
            !,
            S:print('\''),
            data.atom:string(F, Str),
            data.list:iter(Str, Output_char),
            S:print('\''),
            write_args(Args).
         write(number(N)) :-
            !,
            S:print(N).
         write(term(T)) :-
            !,
            S:format('~q', [T]).
         write(T) :-
            throw('write_error'(term(T))).

         write_args([]) :- !.
         write_args([H | T]) :- !,
            S:print('('),
            write(H),
            data.list:iter(T, Output_arg),
            S:print(')').
         write_args(T) :-
            throw('write_error'(args(T))).
      },
      module Output_char [S] {
         do(Char) :-
            data.char:escape(Char, Escape),
            S:print(Escape).
      },
      module Output_arg [S, Writer] {
         do(Arg) :-
            S:print(', '),
            Writer:write(Arg).
      },
      Writer:write(P).

   write_clauses(S, C) :-
      module Write_clause [S] {
         do(Clause) :-
            catch(
               emop.writer:write(S, Clause),
               E,
               throw('write_error'(E, Clause))
            ),
            S:print_endline('.').
      },
      data.list:iter(C, Write_clause).
}
